'''
COMP 116 support module with autograding

Based on the similar module by Gary Bishop for COMP 116 in Fall 2016
We're not doing as much for summer session 2017, though, so this is simpler.
'''

from collections import OrderedDict
import re
import hashlib
import pickle

import matplotlib
import numpy as np

# ----Check functions----
# Checks to see if the given variable is close to ev
def check_array(tag, given, ev, extra):
    '''Compare arrays and array-like things'''
    rtol = 10.0 ** -extra.get('precision', 5)
    if not isinstance(given, np.ndarray):
        try:
            given = np.array(given)
        except:
            print(tag, 'incorrect type, expected array-like')
            return 0.0

    if given.shape != ev.shape:
        print(tag, 'incorrect shape')
        print('  expected', ev.shape, 'got', given.shape)
        return 0.0

    if issubclass(ev.dtype.type, np.number):
        if issubclass(given.dtype.type, np.number):
            if not np.allclose(given, ev, rtol=rtol):
                print(tag, 'incorrect values in array')
                return 0.0
        else:
            print(tag, 'incorrect array type', given.dtype.type)
            print('  expected', ev.dtype.type)
            return 0.0
    
    else:
        try:
            if not np.all(ev == given):
                print(tag, 'incorrect values in array')
                return 0.0
        except:
            print(tag, 'incorrect array value')
            return 0.0

    return 1.0

def check_figure(tag, given, ev, extra):
    '''Compare a few features of figures'''
    rtol = extra.get('relative_tolerance', 1e-5)
    atol = extra.get('absolute_tolerance', 1e-8)
    given = figureState(given)
        
    # compare line graphs
    LineWeight = 0.0
    nlines = len(ev['lines'])
    LOK = np.zeros(nlines)
    if nlines:
        LineWeight = 1.0
        glines = len(given['lines'])
        for i in range(glines):
            gline = given['lines'][i]
            for j in range(nlines):
                if not LOK[j]:
                    eline = ev['lines'][j]
                    LOK[j] = (len(eline[0]) == len(gline[0]) and
                        len(eline[1]) == len(gline[1]) and
                        np.allclose(eline[0], gline[0], rtol=rtol, atol=atol) and
                        np.allclose(eline[1], gline[1], rtol=rtol, atol=atol))
                    if LOK[j]:
                        break

        LineScore = np.mean(LOK)

        if LineScore == 0:
            print(tag, 'values of plotted lines incorrect')
        elif LineScore < 1.0:
            print(tag, 'values of some lines incorrect')
    else:
        LineScore = 0
        
    if LineScore > 0 and glines > nlines:
        print(tag, 'too many lines')
        LineScore *= float(nlines) / glines

    # compare bar charts
    BarWeight = 0.0
    nbars = ev['bars'].shape[1]
    if nbars:
        BarWeight = 1.0
        gbars = given['bars'].shape[1]
        BOK = np.zeros(nbars)
        for i in range(gbars):
            gbar = given['bars'][:,i]
            for j in range(nbars):
                if not BOK[j]:
                    ebar = ev['bars'][:,j]
                    BOK[j] = np.allclose(ebar, gbar, rtol=rtol, atol=atol)
                    if BOK[j]:
                        break
        BarScore = np.mean(BOK)

        if BarScore == 0:
            print(tag, 'bar values incorrect')
        elif BarScore < 1.0:
            print(tag, 'some bar values incorrect')
    else:
        BarScore = 0
        
    # compare labels
    LabelScore = 0.0
    LabelWeight = 0.0
    for attr in ['title', 'xlabel', 'ylabel']:
        if ev[attr]:
            LabelWeight += 1
            if len(given[attr]) == 0:
                print(tag, attr, 'too short')
                print('  consider', ev[attr])
            else:
                LabelScore += 1
    if LabelWeight > 0:
        LabelScore /= LabelWeight
        
        # If LabelWeight is 1/4 of the other weights, that means there are actually 5/4 parts and labels are 1/5 of that
        # This forces labels to always be 20% of the figure score
        LabelWeight = (1/4) * (LineWeight+BarWeight)
        if LabelWeight < 1e-8:
            LabelWeight = 1
        LabelScore /= LabelWeight
        
        
    Score = (LabelScore + LineScore + BarScore) / float(LabelWeight + LineWeight + BarWeight)
    return Score

def check_float(tag, given, ev, extra):
    '''Compare floats for approximate equality'''
    rtol = 10.0 ** -extra.get('precision', 5)
    if not isinstance(given, (float, int)):
        print(tag, 'incorrect type')
        print(' expected float')
        return 0.0
    OK = np.allclose(given, ev, rtol=rtol)
    if not OK:
        print(tag, 'incorrect')
        print('  expected', ev)
    return float(OK)

try:
    # If the solution.py file is in the folder, use those start, check, and report functions
    # That way we can generate the solutions
    from solution import start, check, report, addNonGraded
except ImportError:
    # If there is no solution.py, this is a student's copy, and we need to provide the checker functions instead
    # contains the expected answers
    expected = {}
    
    def start(assignment):
        '''Initialize expected values for checking a student submission'''
        pname = assignment + '.pickle'
        expected.update(pickle.load(open(pname, 'rb')))
        return check, addNonGraded, report
    
    def check(tag, value, **kwargs):
        '''Provide feedback on a single value'''

        assert tag in expected
        e = expected[tag]
            
        dv = describe_answer(value)
        score = 1.0
        
        if e['description'] != dv:
            print(tag, '**incorrect type**')
            print('your answer type is', dv)
            print('expected type is', e['description'])
            score = 0.0
            
        elif 'hash' in e:
            hv = hash_answer(value, kwargs.get('precision', 5))
            if hv != e['hash']:
                score = 0.0
                print(tag, '**incorrect**')
                
        else:
            ev = e['value']
            extra = e['extra']

            if isinstance(ev, np.ndarray):
                score = check_array(tag, value, ev, extra)

            elif isinstance(ev, dict) and 'FigureState' in ev:
                score = check_figure(tag, value, ev, extra)

            elif isinstance(ev, float):
                score = check_float(tag, value, ev, extra)

            elif value == ev:
                pass

            else:
                print(tag, '**incorrect**')
                print('  expected', ev)
                score = 0.0

        if score == 1.0:
            print(tag, 'appears correct')

        elif score > 0:
            print(tag, '*partially correct*')

        e['correct'] = score
        
    def addNonGraded(tag, points):
        return
        
    def tagSort(tags):
        return sorted(tags,
            key=lambda tag: [ s.isdigit() and int(s) or s
                              for s in re.findall(r'\d+|\D+', tag)
                              ])
    
    def report(author, extra):
        '''Summarize the student's performance'''
        expected['_score'] = 0.0
        correct = 0
        ag_problems = 0
        problems = 0
        answered = 0
        ag_points = 0
        max_ag_points = 0
        max_points = 0
        for tag in tagSort(expected.keys()):
            if tag.startswith('_'):
                continue
            
            max_points += expected[tag]['points']
            problems += 1
            
            if 'noauto' in expected[tag]:
                print(tag, 'not auto-graded')
                continue
            
            ag_problems += 1
            c = expected[tag]['correct']
            if c > 0:
                correct += c
                ag_points += expected[tag]['points'] * c
                if c < 1:
                    print(tag, 'partially incorrect')
                else:
                    print(tag, 'appears correct')
            else:
                print(tag, '**incorrect**')
            max_ag_points += expected[tag]['points']
            
        if '_author' in expected and author == expected['_author']:
            print('You must fill in your onyen into the author variable.')
            return
        print("\n----Report for", author, "----")
        
        '''
        if '_exam' in expected and expected['_exam']:
            if not extra:
                print('You must type your name as the value of the pledge variable.')
                return
            else:
                print("  Pledged on my honor:", extra)
                print("   ", getpass.getuser())
        '''

        if '_collaborators' in expected and extra == expected['_collaborators']:
            print('You must fill in the collaborators variable')
            return
        else:
            print("  Collaborators:", extra)
        print("  Autograder: {}/{} possibly correct for up to {}/{} points".format(correct, ag_problems, ag_points, max_ag_points))
        print('  {}/{} Non-autograded problems, accounting for {}/{} points'.format(problems - ag_problems, problems, max_points-max_ag_points, max_points))
        expected['_score'] = ag_points

        return
    
    
# ----Descriptive Functions----
# These functions help describe a variable in words so they can be better understood!
def figureState(f):
    '''Extract interesting bits of the state out of a figure'''
    s = {
        'FigureState':True,
        'title': '',
        'xlabel': '',
        'ylabel': '',
        'lines': [],
        'bars' : np.zeros((3, 0))
    }
    try:
        axis = f.axes[0]
        s['title'] = axis.title.get_text()
        s['xlabel'] = axis.xaxis.label.get_text()
        s['ylabel'] = axis.yaxis.label.get_text()
        s['lines'] = [(line.get_xdata(), line.get_ydata())
                      for line in axis.lines]
        xbars = [p.get_x() for p in axis.patches]
        ybars = [p.get_y() for p in axis.patches]
        hbars = [p.get_height() for p in axis.patches]
        s['bars'] = np.array([xbars, ybars, hbars])
    except:
        pass
    return s

def describe_answer(o):
    '''describe the type of an object in English'''
    def wrap(d):
        '''Enclose description in parenthesis if it contains comma or and.'''
        if ', ' in d or ' and ' in d:
            return '(' + d + ')'
        else:
            return d

    def and_list(items):
        '''comma separated list with and at the end'''
        if len(items) <= 2:
            return " and ".join(items)
        return ", ".join(items[ : -1]) + ", and " + items[-1]

    def describe_sequence(o, typ, memo):
        '''describe a list or tuple'''
        if len(o) == 0:
            return 'empty ' + typ
        if id(o) in memo:
            return '[...]'
        memo.add(id(o))
        et = [ wrap(describe_any(e, memo)) for e in o ]
        uet = list(OrderedDict.fromkeys(et))
        if len(uet) == 1:
            et = '{} {}'.format(len(et), uet[0])
        else:
            et = and_list(et)
        return typ + ' of ' + et
    
    def describe_dict(o, memo):
        if len(o) == 0:
            return 'empty dict'
        if id(o) in memo:
            return '{...}'
        memo.add(id(o))
        it = [ (describe_any(k, memo) + ':' + wrap(describe_any(v, memo))) for k,v in o.items() ]
        uit = list(OrderedDict.fromkeys(it))
        if len(uit) == 1:
            it = '{} {}'.format(len(o), uit[0])
        else:
            it = and_list(it)
        return 'dictionary of {}'.format(it)
    
    def describe_array(a):
        '''Describe a numpy array in English'''
        if issubclass(a.dtype.type, np.integer):
            t = 'integer'
        elif issubclass(a.dtype.type, np.float):
            t = 'float'
        elif issubclass(a.dtype.type, np.complex):
            t = 'complex'
        elif issubclass(a.dtype.type, np.bool_):
            t = 'boolean'
        else:
            t = str(a.dtype)
        s = ' x '.join(str(i) for i in a.shape)
        if s == '0':
            return 'empty array of ' + t
        return 'array of {} {}'.format(s, t)
    
    def describe_plot(o):
        '''describe a plot'''
        o = figureState(o)
        terms = []
        nlines = len(o['lines'])
        if nlines > 0:
            terms.append('{} line{}'.format(nlines, "s"[nlines==1:]))
        nbars = o['bars'].shape[-1]
        if nbars > 0:
            terms.append('{} bar{}'.format(nbars, "s"[nbars==1:]))
        '''
        for t in ['title', 'xlabel', 'ylabel']:
            if o[t]:
                terms.append(t)
        '''
        if len(terms) == 0:
            return 'empty plot'
        else:
            return 'plot with ' + and_list(terms)
        
    def describe_any(o, memo):
        if isinstance(o, str):
            return 'string'
        if isinstance(o, (int, np.integer)):
            return 'integer'
        if isinstance(o, (float, np.floating)):
            return 'float'
        if isinstance(o, (bool, np.bool_)):
            return 'boolean'
        if o is None:
            return 'None'
        if isinstance(o, np.ndarray):
            return describe_array(o)
        if isinstance(o, list):
            return describe_sequence(o, 'list', memo)
        if isinstance(o, tuple):
            return describe_sequence(o, 'tuple', memo)
        if isinstance(o, dict):
            return describe_dict(o, memo)
        if isinstance(o, matplotlib.figure.Figure):
            return describe_plot(o)
        return str(type(o))
    
    desc = describe_any(o, set())
    def pluralize(m):
        n = int(m.group(1))
        w = m.group(3)
        if n != 1:
            if w == 'dictionary':
                w = 'dictionaries'
            else:
                w = w + 's'
        return m.group(1) + ' ' + m.group(2) + w
    desc = re.sub(r'(\d+) (\(?)(tuple|list|dictionary|integer|float|boolean|plot)', pluralize, desc)
    return desc

def hash_answer(o, precision=5):
    '''return a hash to represent the answer in equality tests'''
    def str_answer(o, memo):
        '''compute a hash for an answer'''
        if id(o) in memo:
            return '...'
        memo.add(id(o))
        if isinstance(o, np.ndarray):
            s = 'array' + np.array2string(o, precision=precision).replace('\n', '')
        elif isinstance(o, matplotlib.figure.Figure):
            s = 'figure' + str_answer(sorted(figureState(o).items()), memo)
        elif isinstance(o, float):
            s = 'float' + format(o, '.{}e'.format(precision))
        elif isinstance(o, (list, tuple)):
            s = str(type(o).__name__) + '(' + ','.join([str_answer(i, memo) for i in o]) + ')'
        elif isinstance(o, dict):
            s = 'dict' + str_answer(sorted(o.items()), memo)
        else:
            s = str(o)
        return s

    sa = str_answer(o, set())

    return hashlib.md5(str_answer(o, set()).encode('utf-8')).hexdigest()
